// ---------------------------------------------------------------------------
// Free Pascal Ropes Implementation                         Copyright (C) 2018
// Version 1.00                                                Adriaan Rappard
// ---------------------------------------------------------------------------
// Module     : RopeDemo.pas
// Language   : Free Pascal 3.0.4
// Author     : Adriaan Rappard
// Abstract   : Rope data type implementation - Demo program
// ---------------------------------------------------------------------------
// Comment    : The rope data type is used for efficiently storing and
//              manipulating very long strings. This program demonstrates
//              some of the capabilities of the rope datatatype for text
//              editing.
// ---------------------------------------------------------------------------
// References : https://en.wikipedia.org/wiki/Rope_(data_structure)
// ---------------------------------------------------------------------------
// License    : This program is free software: you can redistribute it and/or
//              modify it under the terms of the GNU General Public License as
//              published by the Free Software Foundation, either version 3 of
//              the License, or (at your option) any later version.
//
//              This program is distributed in the hope that it will be
//              useful, but WITHOUT ANY WARRANTY; without even the implied
//              warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
//              PURPOSE.  See the GNU General Public License for more details.
//
//              You should have received a copy of the GNU General Public
//              License along with this program. If not, see
//              <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
// 2018/10/09 : Adriaan Rappard
//              Initial version
// ---------------------------------------------------------------------------

    {$MODE OBJFPC}

   program RopeDemo;

      // Units used

      uses

         // Console output

         Video,

         // Keyboard input

         Keyboard,

         // Date/time routines

         SysUtils,

         // Basic rope data type implementation

         Ropes;

      // Global constants

      const

         // Maximal unsigned value

         MaxCardinal             = 4294967295;

         // Console dimensions

         ConsoleHeight           = 50;
         ConsoleWidth            = 80;

         // Header and footer text height

         HeaderHeight            = 1;
         FooterHeight            = 1;

         // Content text height

         ContentHeight           = ConsoleHeight-HeaderHeight-FooterHeight;

         // Content scroll amounts

         ContentLineScroll       = 5;
         ContentColumnScroll     = 10;

         // Header text appearance

         HeaderColor             = Black;
         HeaderBackground        = White;

         // Normal content text appearance

         ContentColor            = LightGray;
         ContentBackground       = Blue;

         // Content text appearance

         SelectColor             = Black;
         SelectBackground        = LightGray;

         // Footer text appearance

         FooterColor             = Black;
         FooterBackground        = White;

         // Dialog text apperance

         DialogColor             = Yellow;
         DialogBackground        = Red;

         // Number of slots in history buffer

         HistorySlots            = 25;

         // Space between tab stops

         TabSpace                = 8;

         // Maximum file name length

         MaxNameLength           = 24;

         // Maximum search and replace text length

         MaxStringLength         = 32;

      // Global types

      type

         // Video display color range

         Color                   = 0..16;

         // Text editor state

         State                   = record
            // Textual content
            Content                 : Rope;
            // Cursor position in content
            Position                : Cardinal;
            // Cursor line and column in content
            Line, Column            : Cardinal;
            // Cursor coordinates on CRT
            CursorX, CursorY        : Word;
            // Number of lines and characters in content
            Lines, Characters       : Cardinal;
            // Selected text
            SelectLow, SelectHigh   : Cardinal;
            end; {State}

         // History buffer slot count and index ranges

         HistoryCount            = 0..HistorySlots;
         HistoryIndex            = 0..HistorySlots-1;

         // EDit info settings

         Info                    = (CursorInfo, TextInfo, HistoryInfo,
                                    StorageInfo, NodeInfo);

         // Edit actions

         Action                  = (NullAction, TypeAction, LineAction,
                                    CorrectAction, ScrollAction, SelectAction,
                                    ClipAction, ScanAction);

         // Edit modes

         Mode                    = (InsertMode, OverwriteMode);

      // Global variables

      var

         // Current video display attributes

         Attributes              : Byte;

         // Last event from keyboard

         Event                   : TKeyEvent;

         // Array holding the editor states

         History                 : array [HistoryIndex] of State;

         // Indices of original, current, and most recent state

         OriginSlot,
         CurrentSlot,
         RecentSlot              : HistoryIndex;

         // Clipboard text

         Clipboard               : Rope;

         // Clipboard lines and characters

         ClipLines, ClipChars    : Cardinal;

         // Current edit info setting

         CurrentInfo             : Info;

         // Current edit action

         CurrentAction           : Action;

         // Current edit mode

         CurrentMode             : Mode;

         // Name of file

         FileName                : AnsiString;

         // Search and replace text

         Search, Replace         : AnsiString;

         // Change flag

         TextChanged             : Boolean;

         // Main loop control

         Quit                    : Boolean;

      // Convert cardinal to decimal string

      function CardToStr(Value:Cardinal):AnsiString;

         var

            Factor                  : Cardinal;

            Ch                      : Char;

            R                       : AnsiString;

      begin {CardToStr(Value:Cardinal):AnsiString}
      // Setup factor
      case Value of
         0..9:
            Factor:=1;
         10..99:
            Factor:=10;
         100..999:
            Factor:=100;
         1000..9999:
            Factor:=1000;
         10000..99999:
            Factor:=10000;
         100000..999999:
            Factor:=100000;
         1000000..9999999:
            Factor:=1000000;
         10000000..99999999:
            Factor:=10000000;
         100000000..999999999:
            Factor:=100000000;
         else
            Factor:=1000000000;
         end;
      // Clear result
      R:='';
      // Convert to string
      while Factor<>0 do
         begin
         Ch:=Chr(Ord('0')+Value div Factor);
         Value:=Value mod Factor; Factor:=Factor div 10;
         R:=R+Ch
         end;
      // Return result
      CardToStr:=R
      end; {CardToStr(Value:Cardinal):AnsiString}

      // Count character occurrences in rope

      function Count(R:Rope; Ch:Char):Cardinal;

         var

            // Occurrence counter

            C                       : Cardinal;

            // Loop control variable for position

            K                       : Cardinal;

      begin {Count(R:Rope; Ch:Char):Cardinal}
      // Declare value parameter
      Declare(R);
      // Reset occurrence counter
      C:=0;
      // Loop through rope
      for K:=1 to Length(Load(R)) do
         // Check and count characters
         if Select(Load(R),K)=Ch then C:=C+1;
      // Set result
      Count:=C;
      // Drop value parameter
      Drop(R)
      end; {Count(R:Rope; Ch:Char):Cardinal}

      // Set text color

      procedure TextColor(C:Color);

      begin {TextColor(C:Color)}
      // Set the text color
      Attributes:=(Attributes and %11110000) or C
      end; {TextColor(C:Color)}

      // Set text background color

      procedure TextBackground(C:Color);

      begin {TextBackground(C:Color)}
      // Set the text background color
      Attributes:=(Attributes and %10001111) or ((C and %00000111) shl 4)
      end; {TextBackground(C:Color)}

      // Output text at X,Y position

      procedure TextOutput(X,Y:Word; S:AnsiString);

         var

            // Length of string

            L                       : Cardinal;

            // Loop control variables for index and position

            I, K                    : Word;

      begin {TextOutput(X,Y:Word; S:AnsiString)}
      // Set position in video buffer
      K:=X+Y*ScreenWidth;
      // Set length of string
      L:=Length(S);
      // Check for screen edge and adjust to fit
      if X+L>ScreenWidth-1 then L:=ScreenWidth-X;
      // Copy string into video buffer
      for I:=1 to L do
         VideoBuf^[K+I-1]:=(Attributes shl 8) or Ord(S[I])
      end; {TextOutput(X,Y:Word; S:AnsiString)}

      // Clear to end of the line from X,Y position

      procedure ClearEoLn(X,Y:Word);

         var

            // Loop control variables for index and position

            I, K                    : Word;

      begin {ClearEoLn(X,Y:Word)}
      // Set position in video buffer
      K:=X+Y*ScreenWidth;
      // Fill video buffer with blank space upto the end of the line
      for I:=1 to ScreenWidth-X do
         VideoBuf^[K+I-1]:=(Attributes shl 8) or Ord(' ')
      end; {ClearEoLn(X,Y:Word)}

      // Return undo states count

      function UndoCount:HistoryCount;

      begin {UndoCount}
      // Check if origin state index is less than current state index
      if OriginSlot<=CurrentSlot then
         // No edge wrap
         UndoCount:=CurrentSlot-OriginSlot
      else
         // Edge wrap
         UndoCount:=HistorySlots+CurrentSlot-OriginSlot
      end; {UndoCount}

      // Return redo states count

      function RedoCount:HistoryCount;

      begin {RedoCount}
      // Check if current state index is less than recent state index
      if CurrentSlot<=RecentSlot then
         // No edge wrap
         RedoCount:=RecentSlot-CurrentSlot
      else
         // Edge wrap
         RedoCount:=HistorySlots+RecentSlot-CurrentSlot
      end; {RedoCount}

      // Clear all undo states

      procedure DropUndo;

      begin {DropUndo}
      // Loop through undo states up to current
      while OriginSlot<>CurrentSlot do
         // Select origin state
         with History[OriginSlot] do
            begin
            // Drop content
            Drop(Content);
            // Advance origin to next slot
            OriginSlot:=(OriginSlot+1) mod HistorySlots
            end
      end; {DropUndo}

      // Clear all redo states

      procedure DropRedo;

      begin {DropRedo}
      // Loop through redo states down to current
      while RecentSlot<>CurrentSlot do
         // Select recent state
         with History[RecentSlot] do
            begin
            // Drop content
            Drop(Content);
            // Backup recent to previous slot
            RecentSlot:=(HistorySlots+(RecentSlot-1)) mod HistorySlots
            end
      end; {DropRedo}

      // Get character at position K from current slot

      function GetCh(K:Cardinal):Char;

      begin {GetCh(P:Cardinal):Char}
      // Select current state
      with History[CurrentSlot] do
         // Check if position within content
         if (1<=K) and (K<=Characters) then
            // Position within content, return character
            GetCh:=Select(Load(Content),K)
         else
            // Position outside content, return carriage return
            GetCh:=#010
      end; {GetCh(P:Cardinal):Char}

      // Return line start position of N-th line relative to cursor

      function LinePosition(N:Integer):Cardinal;

         var

            // Loop control variable for position

            K                       : Cardinal;

      begin {LinePosition(N:Integer):Cardinal}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Start at current cursor position within text
         K:=Position;
         // Check N for scan direction
         if N<=0 then
            // Scan backward
            begin
            // Backup to target line
            while N<=0 do
               begin
               // Backup to line terminator
               repeat K:=K-1 until GetCh(K)=#010;
               // Increment line difference N
               N:=N+1
               end;
            LinePosition:=K+1
            end
         else
            // Scan forward
            begin
            // Advance to target line
            while N>0 do
               begin
               // Advance to line terminator
               while GetCh(K)<>#010 do K:=K+1;
               // Advance to next line
               K:=K+1;
               // Decrement line difference N
               N:=N-1
               end;
            // Return first position of target line
            LinePosition:=K
            end
         end
      end; {LinePosition(N:Integer):Cardinal}

      // Hide CRT cursor

      procedure HideCursor;

      begin {HideCursor}
      // Hide cursor
      SetCursorType(crHidden)
      end; {HideCursor}

      // Show CRT cursor

      procedure ShowCursor;

      begin {ShowCursor}
      // Show cursor
      case CurrentMode of
         InsertMode:
            // Insert mode, set underline cursor
            SetCursorType(crUnderline);
         OverwriteMode:
            // Overwrite mode, set block cursor
            SetCursorType(crBlock)
         end
      end; {ShowCursor}

      // Refresh the header text

      procedure RefreshHeader;

         const

            // Header middle CRT row

            Middle                  = (HeaderHeight+1) div 2;

         var

            // Information texts

            CenterText              : AnsiString;

            // CRT Row index

            Y                       : Word;

      begin {RefreshHeader}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Set center information text
         if FileName<>'' then
            // File name set, show file name
            CenterText:=FileName
         else
            // No file name set, show <untitled>
            CenterText:='<untitled>';
         // Set text appearance
         TextColor(HeaderColor); TextBackground(HeaderBackground);
         // Clear header text
         for Y:=0 to HeaderHeight-1 do ClearEoLn(0,Y);
         // Position cursor and write center header info
         TextOutput((ConsoleWidth-Length(CenterText)) div 2,Middle-1,
            CenterText);
         // Refresh video display
         UpdateScreen(False)
         end
      end; {RefreshHeader}

      // Refresh the footer text

      procedure RefreshFooter;

         const

            // Footer middle CRT row

            Middle                  = ConsoleHeight-((FooterHeight-1) div 2);

         var

            // Information texts

            LeftText,
            CenterText,
            RightText               : AnsiString;

            // Number of ropes, concatenations, and leaves

            Nr, Nc, Nl,

            // Allocated storage for ropes, concatenations, and leaves

            Sr, Sc, Sl,

            // Storage utilization percentage

            U                       : Cardinal;

            // CRT Row index

            Y                       : Word;

      begin {RefreshFooter}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Set text appearance
         TextColor(FooterColor); TextBackground(FooterBackground);
         // Clear footer text
         for Y:=ConsoleHeight-FooterHeight to ConsoleHeight-1 do
            ClearEoLn(0,Y);
         // Select information text
         case CurrentInfo of
            CursorInfo:
               begin
               // Set cursor info
               LeftText:='Line:'+CardToStr(Line)+
                  ' Column:'+CardToStr(Column);
               CenterText:='Cursor';
               RightText:='Position:'+CardToStr(Position)
               end;
            TextInfo:
               begin
               // Set text info
               LeftText:='Lines:'+CardToStr(Lines)+
                  ' Characters:'+CardToStr(Characters);
               CenterText:='Text';
               RightText:='Pages:'+CardToStr((Lines div ContentHeight)+1)
               end;
            HistoryInfo:
               begin
               // Set history info
               LeftText:='Undo''s:'+CardToStr(UndoCount)+
                  ' Redo''s:'+CardToStr(RedoCount);
               CenterText:='History';
               RightText:='Total:'+CardToStr(UndoCount+1+RedoCount)
               end;
            StorageInfo:
               begin
               // Compute variable count
               Nr:=RopeVariables;
               // Compute node counts
               Nc:=ConcatNodes; Nl:=LeafNodes;
               // Compute variable storage allocation
               Sr:=Nr*SizeOf(Rope);
               // Compute node storage allocation
               Sc:=Nc*SizeOf(ConcatNode); Sl:=Nl*SizeOf(LeafNode);
               // Compute utilization percentage
               U:=100; if Sr<>0 then U:=Round(100.0*(Length+Sr)/(Sr+Sl+Sc));
               // Set storage info
               LeftText:='Characters:'+CardToStr(Length)+
                  ' Bytes:'+CardToStr(Sr+Sc+Sl);
               CenterText:='Storage';
               RightText:='Utilization:'+CardToStr(U)+'%'
               end;
            NodeInfo:
               begin
               // Compute node counts
               Nc:=ConcatNodes; Nl:=LeafNodes;
               // Set node info
               LeftText:='Concatenations:'+CardToStr(Nc)+
                  ' Leaves:'+CardToStr(Nl);
               CenterText:='Nodes';
               RightText:='Total:'+CardToStr(Nc+Nl)
               end
            end;
         // Position cursor and write left information text
         TextOutput(1,Middle-1,LeftText);
         // Position cursor and write center information text
         TextOutput((ConsoleWidth-Length(CenterText)) div 2,Middle-1,
            CenterText);
         // Position cursor and write right information text
         TextOutput(ConsoleWidth-(Length(RightText)+1),Middle-1,
            RightText);
         // Refresh video display
         UpdateScreen(False)
         end
      end; {RefreshFooter}

      // Refresh the content text

      procedure RefreshContent;

         var

            // Visible lines of content

            Top, Bottom             : Cardinal;

            // Visible columns of content

            Left, Right             : Cardinal;

            // Character selected from content

            Ch                      : Char;

            // Loop control variables for line, column and position

            I, J, K                 : Cardinal;

            // CRT column and row index

            X, Y                    : Word;

      begin {RefreshContent}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Determine visible lines
         Top:=Line-(CursorY-HeaderHeight); Bottom:=Top+ContentHeight;
         // Determine visible columns
         Left:=Column-CursorX; Right:=Left+ConsoleWidth-1;
         // Set content cursor at first position of the top line
         I:=Top; K:=LinePosition(HeaderHeight-CursorY);
         // Set CRT row at top of content window
         Y:=HeaderHeight;
         // Loop through visible lines
         while I<>Bottom do
            begin
            // Set normal colors before clearing CRT row
            TextColor(ContentColor); TextBackground(ContentBackground);
            // Set CRT column at leftmost column and clear CRT row
            X:=0; ClearEoLn(X,Y);
            // Set content column at first position of line
            J:=1;
            // Scan forward up to line terminator
            repeat
               // Get the next character of the line
               Ch:=GetCh(K);
               // Set content text appearance
               if (SelectLow<=K) and (K<=SelectHigh) then
                  begin
                  // Set select colors within selection
                  TextColor(SelectColor); TextBackground(SelectBackground)
                  end
               else
                  begin
                  // Set normal colors outside selection
                  TextColor(ContentColor); TextBackground(ContentBackground)
                  end;
               // Refresh content for visible columns
               if (Left<=J) and (J<=Right) then
                  begin
                  // Write character replacing terminator by space
                  if Ch<>#010 then
                     TextOutput(X,Y,Ch)
                  else
                     TextOutput(X,Y,' ');
                  // Advance CRT column
                  X:=X+1
                  end;
               // Advance content column and cursor
               J:=J+1; K:=K+1
            until Ch=#010;
            // Advance to next line and CRT row
            I:=I+1; Y:=Y+1
            end;
         // Reposition cursor
         SetCursorPos(CursorX,CursorY);
         // Refresh video display
         UpdateScreen(False)
         end
      end; {RefreshContent}

      // Ask confirmation and wait for answer

      function Confirm(Q:AnsiString):Boolean;

         const

            // CRT row index

            Y                       = (ConsoleHeight div 2)-1;

         var

            // CRT column index

            X                       : Word;

            // Answer key entered

            R                       : Char;

      begin {Confirm(Q:AnsiString):Boolean}
      // Compute column index
      X:=(ConsoleWidth-(Length(Q)+2)) div 2;
      // Hide cursor
      HideCursor;
      // Set select colors within selection
      TextColor(DialogColor); TextBackground(DialogBackground);
      // Output question text
      TextOutput(X,Y,' '+Q+' ');
      // Refresh video display
      UpdateScreen(False);
      // Wait for a yes or no answer
      repeat
         // Read keyboard
         R:=GetKeyEventChar(TranslateKeyEvent(GetKeyEvent))
      until R in ['Y','N','y','n'];
      // Show cursor
      ShowCursor;
      // Set result
      Confirm:=R in ['Y','y'];
      // Refresh content
      RefreshContent
      end; {Confirm(Q:AnsiString):Boolean}

      // Enter text string of limited length

      function Enter(Q:AnsiString; L:Word):AnsiString;

         const

            // CRT row index

            Y                       = (ConsoleHeight div 2)-1;

         var

            // CRT column index

            X                       : Word;

            // String entered

            R                       : AnsiString;

            // Character entered

            Ch                      : Char;

      begin {Enter(Q:AnsiString; L:Word):AnsiString}
      // Compute column index
      X:=(ConsoleWidth-(Length(Q)+L+2)) div 2;
      // Hide cursor
      HideCursor;
      // Set select colors within selection
      TextColor(DialogColor); TextBackground(DialogBackground);
      // Output question text
      TextOutput(X,Y,' '+Q+StringOfChar(' ',L+1));
      // Refresh video display
      UpdateScreen(False);
      // Advance column index
      X:=X+Length(Q)+1;
      // Set cursor position
      SetCursorPos(X,Y);
      // Show underline cursor
      SetCursorType(crUnderline);
      // Clear result
      R:='';
      // Get first character
      Ch:=GetKeyEventChar(TranslateKeyEvent(GetKeyEvent));
      // Loop until enter key was pressed
      while Ch<>#013 do
         begin
         // Check character
         case Ch of
            // Regular character
            ' '..'~':
               // Check for maximum length
               if Length(R)<>L then
                  begin
                  // Output character and advance column
                  TextOutput(X,Y,Ch); X:=X+1;
                  // Append character to result
                  AppendStr(R,Ch)
                  end;
            // Backspace
            #008:
               // Check for empty string
               if R<>'' then
                  begin
                  // Backup column and output blank space
                  X:=X-1; TextOutput(X,Y,' ');
                  // Remove character from result
                  R:=LeftStr(R,Length(R)-1);
                  end;
            // Escape
            #027:
               begin
               // Reset result and return
               R:=''; Ch:=#013
               end
            end;
         // Set cursor position
         SetCursorPos(X,Y);
         // Refresh video display
         UpdateScreen(False);
         // Get next character
         if Ch<>#013 then Ch:=GetKeyEventChar(TranslateKeyEvent(GetKeyEvent))
         end;
      // Reset cursor
      ShowCursor;
      // Set result
      Enter:=R;
      // Refresh content
      RefreshContent
      end; {Enter(Q:AnsiString; L:Word):AnsiString}

      // Set current action

      procedure SetAction(A:Action);

         var

            // Previous slot index

            PreviousSlot            : HistoryIndex;

      begin {SetAction(A:Action)}
      // Check for action change
      if A<>CurrentAction then
         begin
         // Clear all redo states
         DropRedo;
         // Keep previous slot index
         PreviousSlot:=CurrentSlot;
         // Advance current and recent to next slot
         CurrentSlot:=(CurrentSlot+1) mod HistorySlots;
         RecentSlot:=CurrentSlot;
         // Check if history buffer full
         if CurrentSlot=OriginSlot then
            begin
            // Drop origin if history buffer full
            Drop(History[OriginSlot].Content);
            // Advance origin to next slot
            OriginSlot:=(OriginSlot+1) mod HistorySlots
            end;
         // Copy state from previous state
         with History[CurrentSlot] do
            begin
            // Create content
            Make(Content);
            // Copy content from previous state
            Store(Content,Load(History[PreviousSlot].Content));
            // Copy content cursor position from previous state
            Position:=History[PreviousSlot].Position;
            // Copy content cursor line and column from previous state
            Line:=History[PreviousSlot].Line;
            Column:=History[PreviousSlot].Column;
            // Copy CRT cursor coordinates from previous state
            CursorX:=History[PreviousSlot].CursorX;
            CursorY:=History[PreviousSlot].CursorY;
            // Copy line and character counts from previous state
            Lines:=History[PreviousSlot].Lines;
            Characters:=History[PreviousSlot].Characters;
            // Copy selection from previous state
            SelectLow:=History[PreviousSlot].SelectLow;
            SelectHigh:=History[PreviousSlot].SelectHigh
            end;
         // Make new action current
         CurrentAction:=A
         end
      end; {SetAction(A:Action)}

      // Move the cursor a number of rows up

      procedure Up(N:Cardinal);

         var

            // Loop control variable for column

            J                       : Cardinal;

      begin {Up(N:Cardinal)}
      // Select current state
      with History[CurrentSlot] do
         // Check for valid line range
         if N<Line then
            begin
            // Move N lines up
            Line:=Line-N;
            // Set position to first column in target line
            Position:=LinePosition(-N); J:=1;
            // Reposition cursor in content
            while (GetCh(Position)<>#010) and (J<>Column) do
               begin
               Position:=Position+1; J:=J+1
               end;
            // Check for top edge
            if CursorY-N<HeaderHeight then
               // Scroll window towards top
               begin
               // Check for low line range
               if Line<HeaderHeight+ContentLineScroll then
                  CursorY:=Line+HeaderHeight-1
               else
                  CursorY:=HeaderHeight+ContentLineScroll-1;
               // Refresh content
               RefreshContent
               end
            else
               CursorY:=CursorY-N;
            // Reposition cursor
            SetCursorPos(CursorX,CursorY);
            // Refresh footer info
            RefreshFooter
            end
      end; {Up(N:Cardinal)}

      // Move the cursor a number of rows down

      procedure Down(N:Cardinal);

         var

            // Loop control variable for column

            J                       : Cardinal;

      begin {Down(N:Cardinal)}
      // Select current state
      with History[CurrentSlot] do
         // Check for valid line range
         if Line<=Lines-N then
            begin
            // Move N lines down
            Line:=Line+N;
            // Set position to first column in target line
            Position:=LinePosition(N); J:=1;
            // Reposition cursor in content
            while (GetCh(Position)<>#010) and (J<>Column) do
               begin
               Position:=Position+1; J:=J+1
               end;
            // Check for bottom edge
            if CursorY+N>HeaderHeight+ContentHeight-1 then
               // Scroll window towards bottom
               begin
               // Reposition cursor
               CursorY:=HeaderHeight+ContentHeight-ContentLineScroll;
               // Refresh content
               RefreshContent
               end
            else
               CursorY:=CursorY+N;
            // Reposition cursor
            SetCursorPos(CursorX,CursorY);
            // Refresh footer info
            RefreshFooter
            end
      end; {Down(N:Cardinal)}

      // Move the cursor a number of columns left

      procedure Left(N:Cardinal);

         var

            // Loop control variable for column

            J                       : Cardinal;

      begin {Left(N:Cardinal)}
      // Select current state
      with History[CurrentSlot] do
         // Check for valid column range
         if N<Column then
            begin
            // Move N columns left
            Column:=Column-N;
            // Set position to first column in current line
            Position:=LinePosition(0); J:=1;
            // Reposition cursor in content
            while (GetCh(Position)<>#010) and (J<>Column) do
               begin
               Position:=Position+1; J:=J+1
               end;
            // Check for left edge
            if CursorX-N<0 then
               // Scroll window to the left
               begin
               // Check for low column range
               if Column<ContentColumnScroll then
                  CursorX:=Column-1
               else
                  CursorX:=ContentColumnScroll-1;
               // Refresh content
               RefreshContent
               end
            else
               CursorX:=CursorX-N;
            // Reposition cursor
            SetCursorPos(CursorX,CursorY);
            // Refresh footer info
            RefreshFooter
            end
      end; {Left(N:Cardinal)}

      // Move the cursor a number of columns right

      procedure Right(N:Cardinal);

         var

            // Loop control variable for column

            J                       : Cardinal;

      begin {Right(N:Cardinal)}
      // Select current state
      with History[CurrentSlot] do
         // Check for valid column range
         if Column<=MaxCardinal-N then
            begin
            // Move N columns right
            Column:=Column+N;
            // Set position to first column in current line
            Position:=LinePosition(0); J:=1;
            // Reposition cursor in content
            while (GetCh(Position)<>#010) and (J<>Column) do
               begin
               Position:=Position+1; J:=J+1
               end;
            // Check for right edge
            if CursorX+N>ConsoleWidth-1 then
               // Scroll window to the right
               begin
               // Reposition cursor
               CursorX:=ConsoleWidth-ContentColumnScroll;
               // Refresh content
               RefreshContent
               end
            else
               // Move cursor to the right
               CursorX:=CursorX+N;
            // Reposition cursor
            SetCursorPos(CursorX,CursorY);
            // Refresh footer info
            RefreshFooter
            end
      end; {Right(N:Cardinal)}

      // Scroll one page up

      procedure PageUp;

         var

            // Loop control variable for column

            J                       : Cardinal;

      begin {PageUp}
      // Select current state
      with History[CurrentSlot] do
         // Check for valid line range
         if ContentHeight<Line then
            begin
            // Scroll one page up
            Line:=Line-ContentHeight;
            // Set position to first column in target line
            Position:=LinePosition(-ContentHeight); J:=1;
            // Reposition cursor in content
            while (GetCh(Position)<>#010) and (J<>Column) do
               begin
               Position:=Position+1; J:=J+1
               end;
            // Check for top edge
            if Line-(CursorY-HeaderHeight)<1 then
               // Adjust cursor
               CursorY:=Line+HeaderHeight-1;
            // Reposition cursor
            SetCursorPos(CursorX,CursorY);
            // Refresh content and footer info
            RefreshContent; RefreshFooter
            end
         else
            // If not already on move up to first line
            if 1<Line then Up(Line-1)
      end; {PageUp}

      // Scroll one page down

      procedure PageDown;

         var

            // Loop control variable for column

            J                       : Cardinal;

      begin {PageDown}
      // Select current state
      with History[CurrentSlot] do
         // Check for valid line range
         if Line<=Lines-ContentHeight then
            begin
            // Scroll one page down
            Line:=Line+ContentHeight;
            // Set position to first column in target line
            Position:=LinePosition(ContentHeight); J:=1;
            // Reposition cursor in content
            while (GetCh(Position)<>#010) and (J<>Column) do
               begin
               Position:=Position+1; J:=J+1
               end;
            // Reposition cursor
            SetCursorPos(CursorX,CursorY);
            // Refresh content and footer info
            RefreshContent; RefreshFooter
            end
         else
            // If not already on move down to last line
            if Line<Lines then Down(Lines-Line)
      end; {PageDown}

      // Move cursor to the start of the current line

      procedure LineStart;

      begin {LineStart}
      // Select current state and move to the leftmost column
      with History[CurrentSlot] do Left(Column-1)
      end; {LineStart}

      // Move cursor to the end of the current line

      procedure LineEnd;

         var

            // Length of the current line

            N                       : Cardinal;

      begin {LineEnd}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Reset position to the start of the line
         Position:=LinePosition(0);
         // Reset content cursor and CRT column
         Column:=1; CursorX:=0;
         // Determine the length of the current line
         N:=0; while GetCh(Position+N)<>#010 do N:=N+1;
         // Move right up to end of current line
         Right(N)
         end
      end; {LineEnd}

      // Undo change

      procedure UndoChange;

      begin {UndoChange}
      // Check if there are changes to undo
      if UndoCount<>0 then
         begin
         // Select previous slot
         CurrentSlot:=(HistorySlots+CurrentSlot-1) mod HistorySlots;
         // Refresh content and footer info
         RefreshContent; RefreshFooter
         end
      end; {UndoChange}

      // Redo change

      procedure RedoChange;

      begin {RedoChange}
      // Check if there are changes to undo
      if RedoCount<>0 then
         begin
         // Select previous slot
         CurrentSlot:=(CurrentSlot+1) mod HistorySlots;
         // Refresh content and footer info
         RefreshContent; RefreshFooter
         end
      end; {RedoChange}

      // Insert blanks until CRT cursor within content

      procedure InsertBlanks;

         var

            // Start position of the current line

            K                       : Cardinal;

      begin {InsertBlanks}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Determine start position of the current line
         K:=LinePosition(0);
         // Insert blanks until CRT cursor within content
         while K+Column-1>Position do
            begin
            // Insert blank after current position within string
            Insert(Load(' '),Content,Position-1);
            // Increment character count
            Characters:=Characters+1;
            // Advance position
            Position:=Position+1
            end
         end
      end; {InsertBlanks}

      // Delete blanks before cursor

      procedure DeleteBlanks;

      begin {DeleteBlanks}
      // Select current state
      with History[CurrentSlot] do
         // Delete blanks before current position
         while GetCh(Position-1)=' ' do
            begin
            // Backup one character
            Position:=Position-1;
            // Delete blank
            Delete(Content,Position,1);
            // Decrement character count
            Characters:=Characters-1
            end
      end; {DeleteBlanks}

      // Delete character at cursor

      procedure DeleteCharacter;

      begin {DeleteCharacter}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Insert blanks until CRT cursor within content
         InsertBlanks;
         // Check for empty content
         if Characters<>0 then
            begin
            // Check for line terminator
            if GetCh(Position)=#010 then
               begin
               // Line terminator, position cursor at the end of the line
               LineEnd;
               // Line terminator, decrement line count
               Lines:=Lines-1
               end;
            // Delete character at current position
            Delete(Content,Position,1);
            // Decrement character count
            Characters:=Characters-1;
            // Delete trailing blanks before line terminator
            if GetCh(Position)=#010 then DeleteBlanks;
            // Refresh content and footer info
            RefreshContent; RefreshFooter
            end
         end
      end; {DeleteCharacter}

      // Delete character before cursor and backup

      procedure Backspace;

      begin {Backspace}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Insert blanks until CRT cursor within content
         InsertBlanks;
         // Check for empty content
         if (Characters<>0) and (Position<>1) then
            begin
            // Check for start of the line
            if Column=1 then
               begin
               // Scroll to end of previous line
               Up(1); LineEnd
               end
            else
               // Backup one character
               Left(1);
            // Check for line terminator
            if GetCh(Position)=#010 then
               begin
               // Line terminator, position cursor at the end of the line
               LineEnd;
               // Line terminator, decrement line count
               Lines:=Lines-1
               end;
            // Delete character at current position
            Delete(Content,Position,1);
            // Decrement character count
            Characters:=Characters-1;
            // Delete trailing blanks before line terminator
            if GetCh(Position)=#010 then DeleteBlanks;
            // Refresh content and footer info
            RefreshContent; RefreshFooter
            end
         end
      end; {Backspace}

      // Insert tab spacing at cursor

      procedure TabSpacing;

      begin {TabSpacing}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Insert blanks until CRT cursor within content
         InsertBlanks;
         // Insert spaces upto new tab stop
         repeat
            // Insert blank space at current position
            Insert(Load(' '),Content,Position-1);
            // Increment character count
            Characters:=Characters+1;
            // Advance cursor
            Right(1)
         until (Column mod TabSpace)=1;
         // Refresh content and footer info
         RefreshContent; RefreshFooter
         end
      end; {TabSpacing}

      // Insert line terminator at cursor

      procedure LineTerminator;

      begin {LineTerminator}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Delete trailing blanks before cursor
         DeleteBlanks;
         // Insert line terminator after current position
         Insert(Load(#010),Content,Position-1);
         // Increment character count
         Characters:=Characters+1;
         // Increment line count
         Lines:=Lines+1;
         // Move to first column of the next line
         LineStart; Down(1);
         // Refresh content and footer info
         RefreshContent; RefreshFooter
         end
      end; {LineTerminator}

      // Process character typed

      procedure ProcessCharacter(Ch:Char);

      begin {ProcessCharacter(Ch:Char)}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Insert blanks until CRT cursor within content
         InsertBlanks;
         // Check edit mode
         case CurrentMode of
            InsertMode:
               // Insert character
               begin
               // Insert character after current position within string
               Insert(Load(Ch),Content,Position-1);
               // Increment character count
               Characters:=Characters+1;
               // Advance cursor
               Right(1)
               end;
            OverwriteMode:
               // Overwrite character
               begin
               // Check for the end of the line
               if GetCh(Position)<>#010 then
                  begin
                  // Delete character if not at end of the line
                  Delete(Content,Position,1);
                  // Insert character after current position within string
                  Insert(Load(Ch),Content,Position-1)
                  end
               else
                  begin
                  // Insert character after current position within string
                  Insert(Load(Ch),Content,Position-1);
                  // Increment character count
                  Characters:=Characters+1
                  end;
               // Advance cursor
               Right(1)
               end
            end;
         // Refresh content and footer info
         RefreshContent; RefreshFooter
         end
      end; {ProcessCharacter(Ch:Char)}

      // Find search text starting from position

      procedure SearchText(P:Cardinal);

         var

            // Target column

            J                       : Cardinal;

      begin {SearchText(P:Cardinal)}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Clear target column
         J:=0;
         // Check if P within content
         if P<=Length(Load(Content)) then
            // Search string in content starting at position
            J:=Pos(Load(Search),Copy(Load(Content),
               P,Length(Load(Content))-(P-1)));
         // Check if search text was found
         if J<>0 then
            begin
            // Set selection to found text
            SelectLow:=J+(P-1); SelectHigh:=SelectLow+Length(Search)-1;
            // Increment lines
            Down(Line+Count(Copy(Load(Content),
               P,SelectLow-P),#010)-Line);
            // Determine target column
            J:=SelectLow-(LinePosition(0)-1);
            // Scroll left to target column
            if J<Column then Left(Column-J);
            // Scroll right to target column
            if Column<J then Right(J-Column);
            // Set content cursor at selection
            Position:=SelectLow;
            // Reset to select action
            CurrentAction:=SelectAction
            end
         else
            begin
            // Remove selection
            SelectLow:=1; SelectHigh:=0
            end;
         // Refresh content and footer info
         RefreshContent; RefreshFooter
         end
      end; {SearchText(P:Cardinal)}

      // Replace search text at current position

      procedure ReplaceText;

      begin {ReplaceText}
      // Select current state
      with History[CurrentSlot] do
         begin
         Delete(Content,Position,Length(Search));
         // Insert clipboard text at position
         Insert(Load(Replace),Content,Position-1);
         // Adjust character count
         Characters:=Length(Load(Content));
         // Remove selection
         SelectLow:=1; SelectHigh:=0
         end;
      // Refresh content and footer info
      RefreshContent; RefreshFooter;
      // Set change flag
      TextChanged:=True
      end; {ReplaceText}

      // Reset editor for new file

      procedure ResetEdit;

      begin {ResetEdit}
      // Initialize history buffer
      with History[0] do
         begin
         // Create an empty rope
         Make(Content);
         // Cursor position at start
         Position:=1;
         // Cursor line and column at start
         Line:=1; Column:=1;
         // Cursor CRT coordinates at upper left corner of content window
         CursorX:=0; CursorY:=Headerheight;
         // Reposition cursor
         SetCursorPos(CursorX,CursorY);
         // Initialize line and character count
         Lines:=1; Characters:=0;
         // Set selection markers
         SelectLow:=1; SelectHigh:=0
         end;
      // Initialize history indexes
      OriginSlot:=0; CurrentSlot:=0; RecentSlot:=0;
      // Initialize current edit action
      CurrentAction:=NullAction;
      // Set empty filename
      FileName:='';
      // Set empty search and replace text
      Search:=''; Replace:='';
      // Reset change flag
      TextChanged:=False;
      // Refresh header and footer info
      RefreshHeader; RefreshFooter;
      // Refresh content
      RefreshContent
      end; {ResetEdit}

      // New file

      procedure NewFile;

         var

            // Confirmation flag

            Ok                      : Boolean;

      begin {NewFile}
      // Set confirmation flag
      Ok:=True;
      // Text changed, ask confirmation
      if TextChanged then Ok:=Confirm('Text changed, continue?');
      // Check confirmation
      if Ok then
         begin
         // Drop undo states when text was changed
         if TextChanged then DropUndo;
         // Drop redo and current state
         DropRedo; Drop(History[CurrentSlot].Content);
         // Reset editor for new file
         ResetEdit
         end
      end; {NewFile}

      // Save content text in a file

      procedure SaveFile;

         var

            // Text file handle

            Handle                  : Text;

            // Termination flag

            Quit                    : Boolean;

            // Character selected to be written

            Ch                      : Char;

            // Loop control variable for position

            K                       : Cardinal;

      begin {SaveFile}
      // Select current state
      with History[CurrentSlot] do
         begin
         // Check for empty filename
         if FileName='' then
            begin
            // Retrieve filename
            FileName:=Enter('Enter filename:',MaxNameLength);
            // Refresh header info
            RefreshHeader
            end;
         // Set termination flag and position
         Quit:=FileName=''; K:=1;
         // Loop until text saved or aborted
         while not Quit do
            // Catch IO errors
            try
               // Set file name
               Assign(Handle,FileName);
               // Create new or overwrite existing file
               ReWrite(Handle);
               // Write text until
               while (K<=Length(Load(Content))) and not Quit do
                  begin
                  // Select character from text content
                  Ch:=Select(Load(Content),K);
                  // Catch IO errors
                  try
                     if Ch=#010 then WriteLn(Handle) else Write(Handle,Ch);
                  except
                     on EInOutError do
                        Quit:=Confirm('Error writing file. Abort?')
                     end;
                  // Advance position
                  K:=K+1
                  end;
               // Close the file
               Close(Handle);
               // Terminate loop
               Quit:=True
            except
               on EInOutError do
                  // File not created, ask confirmation to try again
                  Quit:=not Confirm('File could not be created. Try again?')
               end;
         // Check if file was saved
         if K>Length(Load(Content)) then
             begin
            // Drop undo states
            DropUndo;
            // Refresh footer info
            RefreshFooter;
            // Reset change flag
            TextChanged:=False
            end
         end
      end; {SaveFile}

      // Read content text from a file

      function OpenFile:Boolean;

         var

            // Text file handle

            Handle                  : Text;

            // Confirmation, termination, and read ready flag

            Ok, Quit, Ready         : Boolean;

            // Character read from file

            Ch                      : Char;

      begin {OpenFile:Boolean}
      // Set confirmation flag
      Ok:=True;
      // Text changed, ask confirmation
      if TextChanged then Ok:=Confirm('Text changed, continue?');
      // Check confirmation
      if Ok then
         begin
         // Drop undo and redo states
         DropUndo; DropRedo;
         // Drop current state
         Drop(History[CurrentSlot].Content);
         // Reset editor for new file
         ResetEdit;
         // Retrieve filename
         FileName:=Enter('Enter filename:',MaxNameLength);
         // Refresh header info
         RefreshHeader;
         // Set termination and read ready flag
         Quit:=FileName=''; Ready:=False;
         // Loop until text read or aborted
         while not Quit do
            // Catch IO errors
            try
               // Set file name
               Assign(Handle,FileName);
               // Create new or overwrite existing file
               Reset(Handle);
               // Read text content from file until end of file
               while not (Eof(Handle) or Quit) do
                  // Catch IO errors
                  try
                     // Read character from file
                     Read(Handle,Ch);
                     // Ignore carriage return characters
                     if Ch<>#013 then
                        // Select origin state
                        with History[0] do
                           begin
                           // Append character to text content
                           Store(Content,Load(Content)+Load(Ch));
                           // Increment characters
                           Characters:=Characters+1;
                           // Line terminator, increment lines
                           if Ch=#010 then Lines:=Lines+1
                           end
                  except
                     on EInOutError do
                        Quit:=Confirm('Error reading file. Abort?')
                     end;
               // Check if ready
               Ready:=Eof(Handle);
               // Close the file
               Close(Handle);
               // Terminate loop
               Quit:=True
            except
               on EInOutError do
                  // File not created, ask confirmation to try again
                  Quit:=not Confirm('File could not be opened. Try again?')
               end;
         // Check if file read succesfully
         if not Ready then
            // File was not read successfully
            begin
            // Drop current state
            Drop(History[CurrentSlot].Content);
            // Reset editor for new file
            ResetEdit
            end
         else
            // File was read succesfully
            begin
            // Refresh content and footer info
            RefreshContent; RefreshFooter
            end
         end
      end; {OpenFile}

      // Set up editor

      procedure SetUp;

      begin {SetUp}
      // Set console dimensions
      ScreenHeight:=ConsoleHeight; ScreenWidth:=ConsoleWidth;
      // Initilize video and keyboard driver
      InitVideo; InitKeyboard;
      // Set content text appearance
      TextColor(ContentColor); TextBackground(ContentBackground);
      // Clear the screen
      ClearScreen;
      // Initialize information setting
      CurrentInfo:=CursorInfo;
      // Set insert mode and underline cursor
      CurrentMode:=InsertMode; SetCursorType(crUnderLine);
      // Create empty clipboard
      Make(Clipboard); ClipLines:=0; ClipChars:=0;
      // Reset editor for new file
      ResetEdit
      end; {Setup}

      // Clean up editor

      procedure CleanUp;

      begin {CleanUp}
      // Reset text appearance
      TextColor(LightGray); TextBackground(Black);
      // Reset cursor
      SetCursorPos(0,0);
      // Clear the screen
      ClearScreen;
      // Drop history
      DropUndo; DropRedo; Drop(History[CurrentSlot].Content);
      // Drop clipboard
      Drop(Clipboard);
      // Deactivate keyboard and video driver
      DoneKeyboard; DoneVideo
      end; {CleanUp}

   begin {RopeDemo}
   // Set up editor
   SetUp;
   // Main loop
   Quit:=False;
   repeat
      // Read keyboard
      Event:=TranslateKeyEvent(GetKeyEvent);
      case GetKeyEventFlags(Event) of
         kbFnKey:
            case GetKeyEventCode(Event) of
               kbdLeft:
                  // Left arrow (Cursor left)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move cursor left one position
                  Left(1)
                  end;
               kbdRight:
                  // Right arrow (Cursor right)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move cursor right one position
                  Right(1)
                  end;
               kbdUp:
                  // Up arrow (Cursor up)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move cursor up one position
                  Up(1)
                  end;
               kbdDown:
                  // Down arrow (Cursor down)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move cursor down one position
                  Down(1)
                  end;
               kbdHome:
                  // Home (Move to start of line)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move to the start of the current line
                  LineStart
                  end;
               kbdEnd:
                  // End (Move to end of the current line)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move to the end of the current line
                  LineEnd
                  end;
               kbdPgUp:
                  // Page up (Previous page)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move to the previous page
                  PageUp
                  end;
               kbdPgDn:
                  // Page down (Next page)
                  begin
                  // Switch to scrolling action
                  SetAction(ScrollAction);
                  // Move to the next page
                  PageDown
                  end;
               kbdInsert:
                  // Insert (Toggle insert/overwrite)
                  begin
                  // Check current mode
                  case CurrentMode of
                     InsertMode:
                        // Set overwrite mode
                        CurrentMode:=OverwriteMode;
                     OverwriteMode:
                        // Set insert mode
                        CurrentMode:=InsertMode
                     end;
                  // Show cursor
                  ShowCursor
                  end;
               kbdDelete:
                  // Delete (Delete character)
                  begin
                  // Switch to correcting action
                  SetAction(CorrectAction);
                  // Delete character at current position
                  DeleteCharacter;
                  // Set change flag
                  TextChanged:=True
                  end
               end;
         kbPhys:
            case Chr(($FF00 and GetKeyEventCode(Event)) shr 8) of
                     // Alt W (Save file as)
               #017: begin
                     // Reset filename and save file
                     FileName:=''; SaveFile
                     end;
                     // Alt E (End selection)
               #018: begin
                     // Switch to selecting action
                     SetAction(SelectAction);
                     // Set end selection marker
                     with History[CurrentSlot] do SelectHigh:=Position;
                     // Refresh content
                     RefreshContent
                     end;
                     // Alt R (Find and replace text)
               #019: begin
                     // Enter text to search for
                     Search:=Enter('Search for:',MaxStringLength);
                     // Check for empty search text
                     if Search<>'' then
                        begin
                        // Switch to scan text action
                        SetAction(ScanAction);
                        // Find search text starting from current position
                        SearchText(History[CurrentSlot].Position);
                        // Check if found
                        if CurrentAction=SelectAction then
                           begin
                           // Enter text to replace with
                           Replace:=Enter('Replace with:',MaxStringLength);
                           // Check for empty replace text
                           if Replace<>'' then
                              begin
                              // Switch to typing action
                              SetAction(TypeAction);
                              // Replace text
                              ReplaceText
                              end
                           end
                        end
                     end;
                     // Alt U (Undo change)
               #022: begin
                     // Backup to previous state
                     UndoChange;
                     // Refresh footer history info
                     RefreshFooter
                     end;
                     // Alt O (Open file)
               #024: OpenFile;
                     // Alt P (Paste from clipboard)
               #025: begin
                     // Switch to typing action
                     SetAction(TypeAction);
                     // Select current state
                     with History[CurrentSlot] do
                        begin
                        // Insert blanks until CRT cursor within content
                        InsertBlanks;
                        // Insert clipboard text at position
                        Insert(Load(Clipboard),Content,Position-1);
                        // Adjust line count
                        Lines:=Lines+ClipLines;
                        // Adjust character count
                        Characters:=Characters+ClipChars
                        end;
                     // Reset to clipboard action
                     CurrentAction:=ClipAction;
                     // Refresh content and footer info
                     RefreshContent; RefreshFooter;
                     // Set change flag
                     TextChanged:=True
                     end;
                     // Alt A (Find again)
                     // Check for empty search text
               #030: if Search<>'' then
                        begin
                        // Switch to scan text action
                        SetAction(ScanAction);
                        // Find search text starting from current position
                        SearchText(History[CurrentSlot].Position+1);
                        // Check if found and replace not empty
                        if (CurrentAction=SelectAction) and (Replace<>'') then
                           // Ask confirmation
                           if Confirm('Replace text?') then
                              begin
                              // Switch to typing action
                              SetAction(TypeAction);
                              // Replace text
                              ReplaceText
                              end
                        end;
                     // Alt S (Save file)
               #031: SaveFile;
                     // Alt D (Redo change)
               #032: begin
                     // Advance to next state
                     RedoChange;
                     // Refresh footer info
                     RefreshFooter
                     end;
                     // Alt F (Find text)
               #033: begin
                     // Enter text to search for
                     Search:=Enter('Search for:',MaxStringLength);
                     // Check for empty search text
                     if Search<>'' then
                        begin
                        // Clear replacement text
                        Replace:='';
                        // Switch to scan text action
                        SetAction(ScanAction);
                        // Find search text starting from current position
                        SearchText(History[CurrentSlot].Position)
                        end
                     end;
                     // Alt X (Exit program)
               #045: if TextChanged then
                        // Text changed, ask confirmation
                        Quit:=Confirm('Text changed, continue?')
                     else
                        // Terminate program
                        Quit:=True;
                     // Alt C (Copy to clipboard)
               #046: begin
                     // Switch to clipboard action
                     SetAction(ClipAction);
                     // Select current state
                     with History[CurrentSlot] do
                        begin
                        // Clear clipboard lines and characters
                        ClipLines:=0; ClipChars:=0;
                        // Check for valid selection
                        if SelectLow<=SelectHigh then
                           begin
                           // Selection valid, copy selection to clipboard
                           Store(Clipboard,Copy(Load(Content),SelectLow,
                              (SelectHigh-SelectLow)+1));
                           // Determine number of lines
                           ClipLines:=Count(Load(Clipboard),#010);
                           // Determine number of characters
                           ClipChars:=Length(Load(Clipboard));
                           // Remove selection
                           SelectHigh:=SelectLow-1
                           end
                        else
                           // Invalid selection, clear clipboard
                           Store(ClipBoard,Empty);
                        // Refresh footer info
                        RefreshContent; RefreshFooter
                        end
                     end;
                     // Alt B (Begin selection)
               #048: begin
                     // Switch to selecting action
                     SetAction(SelectAction);
                     // Set begin selection marker
                     with History[CurrentSlot] do SelectLow:=Position;
                     // Refresh content
                     RefreshContent
                     end;
                     // Alt N (New file)
               #049: NewFile;
                     // Alt 1 (Show cursor info)
               #120: begin
                     CurrentInfo:=CursorInfo; RefreshFooter
                     end;
                     // Alt 2 (Show text info)
               #121: begin
                     CurrentInfo:=TextInfo; RefreshFooter
                     end;
                     // Alt 3 (Show history info)
               #122: begin
                     CurrentInfo:=HistoryInfo; RefreshFooter
                     end;
                     // Alt 4 (Show storage info)
               #123: begin
                     CurrentInfo:=StorageInfo; RefreshFooter
                     end;
                     // Alt 5 (Show node info)
               #124: begin
                     CurrentInfo:=NodeInfo; RefreshFooter
                     end
            end;
         kbASCII:
            case GetKeyEventChar(Event) of
                     // Backspace
               #008: begin
                     // Switch to correcting action
                     SetAction(CorrectAction);
                     // Backup and delete character
                     Backspace;
                     // Set change flag
                     TextChanged:=True
                     end;
                     // Tab
               #009: begin
                     // Switch to typing action
                     SetAction(TypeAction);
                     // Insert tab spacing at cursor
                     TabSpacing;
                     // Set change flag
                     TextChanged:=True
                     end;
                     // Enter (New line)
               #013: begin
                     // Switch to line action
                     SetAction(LineAction);
                     // Insert line terminator at cursor
                     LineTerminator;
                     // Set change flag
                     TextChanged:=True
                     end
               else
                  begin
                  // Switch to typing action
                  SetAction(TypeAction);
                  // Insert or overwrite character at cursor
                  ProcessCharacter(GetKeyEventChar(Event));
                  // Set change flag
                  TextChanged:=True
                  end;
               end
         end
   until Quit;
   // Clean up editor
   CleanUp
   end. {RopeDemo}
