// ---------------------------------------------------------------------------
// Free Pascal Ropes Implementation                         Copyright (C) 2018
// Version 1.00                                                Adriaan Rappard
// ---------------------------------------------------------------------------
// Module     : Ropes.pas
// Language   : Free Pascal 3.0.4
// Author     : Adriaan Rappard
// Abstract   : Rope data type implementation - An alternative for strings
// ---------------------------------------------------------------------------
// Comment    : The rope data type is used for efficiently storing and
//              manipulating very long strings. This unit defines the rope
//              data type and implements routines to handle rope objects.
// ---------------------------------------------------------------------------
// References : https://en.wikipedia.org/wiki/Rope_(data_structure)
// ---------------------------------------------------------------------------
// License    : This program is free software: you can redistribute it and/or
//              modify it under the terms of the GNU General Public License as
//              published by the Free Software Foundation, either version 3 of
//              the License, or (at your option) any later version.
//
//              This program is distributed in the hope that it will be
//              useful, but WITHOUT ANY WARRANTY; without even the implied
//              warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
//              PURPOSE.  See the GNU General Public License for more details.
//
//              You should have received a copy of the GNU General Public
//              License along with this program. If not, see
//              <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
// 2018/10/09 : Adriaan Rappard
//              Initial version
// ---------------------------------------------------------------------------

   {$MODE OBJFPC}

   unit Ropes;

   interface

      // Exposed constants

      const

         // Maximum length of leaf text

         {$IF DEFINED(DEBUG)}

         MaxLLen                 = 16;

         {$ELSE}

         MaxLLen                 = 256;

         {$ENDIF}

         // Maximum allowed depth in rope tree

         MaxLevel                = 64;

      // --- Exposed types ---

      type

         // Node level

         Level                   = 0..MaxLevel;

         // Pointer to generic node

         PGeneric                = ^GenericNode;

         // Pointer to leaf node

         PLeaf                   = ^LeafNode;

         // Pointer to concatenation node

         PConcat                 = ^ConcatNode;

         // Rope type

         Rope                    = array [0..0] of PGeneric;

         // Generic node

         GenericNode             = record
            // Type tag and reference count
            TRCount                 : Cardinal
            end; {GenericNode}

         // Rope leaf node

         LeafNode                = record
            // Type tag and reference count
            TRCount                 : Cardinal;
            // Leaf length: number of characters in leaf text
            LLen                    : 0..MaxLLen;
            // Leaf text character buffer
            LText                   : array [1..MaxLLen] of Char
            end; {LeafNode}

         // Rope concatenation node

         ConcatNode              = record
            // Type tag and reference count
            TRCount                 : Cardinal;
            // Node level within rope tree
            NLevel                  : Level;
            // Node weight: length of text in left branch
            NWeight                 : Cardinal;
            // Left and right branches
            Left, Right             : Rope
            end; {ConcatNode}

         // Rope enumerator

         REnumerator             = class
            private
               // Back reference to rope
               BRef                 : Rope;
               // Current position and rope length
               Pos, Len             : Cardinal;
               // Get character at current position
               function GetCurrent:Char;
            public
               // Construct enumerator
               constructor Create(R:Rope);
               // Destruct enumerator
               destructor Destroy; override;
               // Character at current position
               property Current:Char read GetCurrent;
               // Advance to next position
               function MoveNext:Boolean;
            end; {REnumerator}

      // Exposed constants

      const

         // Empty rope

         Empty                   : Rope = (Nil);

      // --- Functional core routines ---

      // Create empty rope variable

      procedure Make(var R:Rope);

      // Drop rope variable

      procedure Drop(var R:Rope);

      // Declare rope value parameter

      procedure Declare(var R:Rope);

      // Pass function result

      procedure Pass(var R:Rope);

      // Allocate rope variable and retrieve value

      function Load(var R:Rope):Rope;

      // Create rope value from ANSI string

      function Load(Str:AnsiString):Rope;

      // Release rope variable and store new value

      procedure Store(var R:Rope; S:Rope);

      // Select and return value of R[I]

      function Select(R:Rope; I:Cardinal):Char;

      // Concatenate ropes S and T

      operator + (S, T:Rope) R:Rope;

      // Return true when rope S equal T

      operator = (S, T:Rope) R:Boolean;

      // Return true when rope S unequal T

      operator <> (S, T:Rope) R:Boolean;

      // Return true when rope S strictly less than T

      operator < (S, T:Rope) R:Boolean;

      // Return true when rope S strictly greater than T

      operator > (S, T:Rope) R:Boolean;

      // Return true when rope S less than or equal T

      operator <= (S, T:Rope) R:Boolean;

      // Return true when rope S greater than or equal T

      operator >= (S, T:Rope) R:Boolean;

      // Return length of rope R

      function Length(R:Rope):Cardinal;

      // Return length of string S

      function Length(S:AnsiString):Cardinal;

      // Return position of first occurrence of rope S in rope T

      function Pos(S, T:Rope):Cardinal;

      // Concatenate ropes R1 and R2

      function Concat(R1, R2:Rope):Rope;

      // Concatenate ropes R1, R2 and R3

      function Concat(R1, R2, R3:Rope):Rope;

      // Concatenate ropes R1, R2, R3 and R4

      function Concat(R1, R2, R3, R4:Rope):Rope;

      // Return rope of L characters from R starting at R[I]

      function Copy(R:Rope; I, L:Cardinal):Rope;

      // Delete L characters from R starting at R[I]

      procedure Delete(var R:Rope; I, L:Cardinal);

      // Insert rope S in rope R after R[I]

      procedure Insert(S:Rope; var R:Rope; I:Cardinal);

      // Return conversion of R to upper case

      function Upcase(R:Rope):Rope;

      // Return conversion of R to lower case

      function Lowercase(R:Rope):Rope;

      // Read rope text from standard input

      procedure ReadRope(var R:Rope);

      // Read rope text and line terminator from standard input

      procedure ReadLnRope(var R:Rope);

      // Read rope text from text file

      procedure ReadRope(var T:Text; var R:Rope);

      // Read rope text and line terminator from text file

      procedure ReadLnRope(var T:Text; var R:Rope);

      // Write rope text to standard output

      procedure WriteRope(R:Rope);

      // Write rope text and line terminator to standard output

      procedure WriteLnRope(R:Rope);

      // Write rope text to text file

      procedure WriteRope(var T:Text; R:Rope);

      // Write rope text and line terminator to text file

      procedure WriteLnRope(var T:Text; R:Rope);

      // --- Support and diagnostic routines ---

      // Enumerator operator to allow for .. in loops

      operator enumerator (R:Rope):REnumerator;

      // Return rope leaf node count for rope R

      function LeafNodes(R:Rope):Cardinal;

      // Return rope concatenation node count for rope R

      function ConcatNodes(R:Rope):Cardinal;

      // Return rope tree depth for rope R

      function Depth(R:Rope):Level;

      // Return current rope variable count

      function RopeVariables:Cardinal;

      // Return total leaf node count for current rope variables

      function LeafNodes:Cardinal;

      // Return concatenation node count for current rope variables

      function ConcatNodes:Cardinal;

      // Return total length of current rope variables

      function Length:Cardinal;

      // --- Debug routines ---

      {$IF DEFINED(DEBUG)}

      // Test split rope R at index I in S and T

      procedure TestSplit(var R:Rope; I:Cardinal; var S, T:Rope);

      // List structure of rope

      procedure List(R:Rope);

      {$ENDIF}

   implementation

      uses

         // Exception handling

         Sysutils;

      // --- Private variables ---

      var

         // Number of current rope variables

         RCount                  : Cardinal = 0;

         // Number of leaf nodes allocated

         LCount                  : Cardinal = 0;

         // Number of concatenation nodes allocated

         CCount                  : Cardinal = 0;

         // Total length of current rope variables

         RLength                 : Cardinal = 0;

      // --- Private routines ---

      // Create new empty leaf node

      procedure MakeLeaf(var R:Rope);

      begin {MakeLeaf(var R:Rope)}
      // Allocate memory
      New(PLeaf(R[0]));
      // Initialize leaf node contents
      with PLeaf(R[0])^ do
         begin
         // Clear tag and set reference count
         TRCount:=$00000001;
         // Clear length
         LLen:=0
         end;
      // Check for leaf node counter overflow
      if LCount=$FFFFFFFF then
         raise ERangeError.Create('Leaf node counter overflow');
      // Increment leaf node count
      LCount:=LCount+1
      end; {MakeLeaf(var R:Rope)}

      // Drop leaf node

      procedure DropLeaf(var R:Rope);

      begin {DropLeaf(var R:Rope)}
      // Release memory and clear variable
      Dispose(PLeaf(R[0])); R[0]:=Nil;
      // Check for leaf node counter zero
      if LCount=$00000000 then
         raise ERangeError.Create('Leaf node counter zero');
      // Keep track of allocated nodes
      LCount:=LCount-1
      end; {DropLeaf(var R:Rope)}

      // Create new concatenation node concatenating S and T

      procedure MakeConcat(var R:Rope; S, T: Rope; L:Level);

      begin {MakeConcat(var R:Rope; S, T: Rope; L:Level)}
      // Allocate memory
      New(PConcat(R[0]));
      // Initialize concatenation node contents
      with PConcat(R[0])^ do
         begin
         // Set tag and reference count
         TRCount:=$80000001;
         // Set level
         NLevel:=L;
         // Compute weight
         NWeight:=Length(Load(S));
         // Set branches
         Left:=S; Right:=T
         end;
      // Check for concatenation node counter overflow
      if CCount=$FFFFFFFF then
         raise ERangeError.Create('Concatenation node counter overflow');
      // Increment concatenation node count
      CCount:=CCount+1
      end; {MakeConcat(var R:Rope; S, T: Rope; L:Level)}

      // Drop concatenation node

      procedure DropConcat(var R:Rope);

      begin {DropConcat(var R:Rope)}
      // Release memory and clear variable
      Dispose(PConcat(R[0])); R[0]:=Nil;
      // Check for concatenation node counter zero
      if CCount=$00000000 then
         raise ERangeError.Create('Concatenation node counter zero');
      // Keep track of allocated nodes
      CCount:=CCount-1
      end; {DropConcat(var R:Rope)}

      // Allocate rope variable

      function Allocate(R:Rope):Rope;

      begin {Allocate(R:Rope):Rope}
      // Check for empty rope
      if R[0]<>Nil then
         with R[0]^ do
            begin
            // Check for reference counter overflow
            if (TRCount and $7FFFFFFF)=$7FFFFFFF then
               raise ERangeError.Create('Reference counter overflow');
            // Increment reference count
            TRCount:=TRCount+1
            end;
      // Return rope value
      Allocate:=R
      end; {Allocate(R:Rope):Rope}

      // Release rope variable

      procedure Release(var R:Rope);

      begin {Release(var R:Rope)}
      // Check for empty rope
      if R[0]<>Nil then
         with R[0]^ do
            begin
            // Check for reference counter zero
            if (TRCount and $7FFFFFFF)=0 then
               raise ERangeError.Create('Reference counter zero');
            // Decrement refrence count
            TRCount:=TRCount-1;
            if (TRCount and $7FFFFFFF)=0 then
               begin
               case TRCount and $80000000 of
                  $00000000:
                     // Rope leaf
                     // Drop leaf node
                     DropLeaf(R);
                  $80000000:
                     // Rope concatenation
                     with PConcat(R[0])^ do
                        begin
                        // Release left and right branch
                        Release(Left); Release(Right);
                        // Drop concatenation node
                        DropConcat(R)
                        end
                  end;
               // Return empty rope
               R[0]:=Nil
               end
            end
      end; {Release(var R:Rope)}

      // Construct enumerator

      constructor REnumerator.Create(R:Rope);

      begin {REnumerator.Create(R:Rope)}
      // Initialize instance
      BRef:=R; Pos:=0; Len:=Length(Load(R))
      end; {REnumerator.Create(R:Rope)}

      // Destruct enumerator

      destructor REnumerator.Destroy;

      begin {REnumerator.Destroy}
      // Release back reference
      Release(BRef);
      // Desroy instance
      inherited Destroy
      end; {REnumerator.Destroy}

      // Get character at current position

      function REnumerator.GetCurrent:Char;

      begin {REnumerator.GetCurrent:Char}
      // Return position
      GetCurrent:=Select(Load(BRef),Pos)
      end;{REnumerator.GetCurrent:Char}

      // Advance to next position

      function REnumerator.MoveNext:Boolean;

      begin {REnumerator.MoveNext:Boolean}
      // Advance position
      Pos:=Pos+1;
      // Return false if past end of rope
      MoveNext:=Pos<=Len
      end; {REnumerator.MoveNext:Boolean}

      // Rebalance rope variable

      procedure Rebalance(var R:Rope);

         var

            S                       : Rope;

            Dl, Dr                  : Level;

      begin {Rebalance(var R:Rope)}
      // Rebalancing is called for concatenations only
      with PConcat(R[0])^ do
         begin
         // Determine left and right depth
         Dl:=Depth(Load(Left)); Dr:=Depth(Load(Right));
         // Check rebalance condition
         if Abs(Dl-Dr)>1 then
            // Rebalancing is required
            if Dl<Dr then
               // Rotate left
               begin
               // Rotate top and right node
               S:=Right; Right:=PConcat(S[0])^.Left; PConcat(S[0])^.Left:=R;
               // Recompute weight
               PConcat(S[0])^.NWeight:=Length(Load(PConcat(S[0])^.Left));
               // Adjust depth of top node (now on left)
               with PConcat(PConcat(S[0])^.Left[0])^ do
                  begin
                  // Determine left and right depth
                  Dl:=Depth(Load(Left)); Dr:=Depth(Load(Right));
                  // Recompute level
                  if Dl<Dr then NLevel:=Dr+1 else NLevel:=Dl+1
                  end;
               // Adjust depth of right node (now on top)
               with PConcat(S)^ do
                  begin
                  // Determine left and right depth
                  Dl:=Depth(Load(Left)); Dr:=Depth(Load(Right));
                  // Recompute level
                  if Dl<Dr then NLevel:=Dr+1 else NLevel:=Dl+1
                  end;
               // Return new top
               R:=S
               end
            else
               // Rotate right
               begin
               // Rotate top and left node
               S:=Left; Left:=PConcat(S[0])^.Right; PConcat(S[0])^.Right:=R;
               // Recompute weight
               PConcat(R[0])^.NWeight:=Length(Load(PConcat(R[0])^.Left));
               // Adjust depth of top node (now on right)
               with PConcat(PConcat(S[0])^.Right[0])^ do
                  begin
                  // Determine left and right depth
                  Dl:=Depth(Load(Left)); Dr:=Depth(Load(Right));
                  // Recompute level
                  if Dl<Dr then NLevel:=Dr+1 else NLevel:=Dl+1
                  end;
               // Adjust depth of left node (now on top)
               with PConcat(S)^ do
                  begin
                  // Determine left and right depth
                  Dl:=Depth(Load(Left)); Dr:=Depth(Load(Right));
                  // Recompute level
                  if Dl<Dr then NLevel:=Dr+1 else NLevel:=Dl+1
                  end;
               // Return new top
               R:=S
               end
         end
      end; {Rebalance(var R:Rope)}

      // Split rope R at index I in S and T

      procedure Split(var R:Rope; I:Cardinal; var S, T:Rope);

         var

            L                       : Level;

            J                       : Cardinal;

      begin {Split(var R:Rope; I:Cardinal; var S, T:Rope)}
      // Check for empty rope
      if R[0]<>Nil then
         case R[0]^.TRCount and $80000000 of
            $00000000:
               // Rope leaf
               // Check index against bounds
               if (0<I) and (I<PLeaf(R[0])^.LLen) then
                  if (PLeaf(R[0])^.TRCount and $7FFFFFFF)<>1 then
                     // Leaf is referenced, copy and split leaf
                     begin
                     // Create new leaf for left part
                     MakeLeaf(S);
                     // Copy left part into new leaf
                     for J:=1 to I do
                        // Copy next item from leaf
                        with PLeaf(S[0])^ do
                           begin
                           LLen:=LLen+1;
                           LText[LLen]:=PLeaf(R[0])^.LText[J]
                           end;
                     // Create new leaf for right part
                     MakeLeaf(T);
                     // Copy right part into new leaf
                     for J:=I+1 to PLeaf(R[0])^.LLen do
                        // Copy next item from leaf
                        with PLeaf(T[0])^ do
                           begin
                           LLen:=LLen+1;
                           LText[LLen]:=PLeaf(R[0])^.LText[J]
                           end
                     end
                  else
                     // Leaf is not referenced, just split leaf
                     begin
                     // Create new leaf for right part
                     MakeLeaf(T);
                     // Copy right part into new leaf
                     for J:=I+1 to PLeaf(R[0])^.LLen do
                        // Copy next item from leaf
                        with PLeaf(T[0])^ do
                           begin
                           LLen:=LLen+1;
                           LText[LLen]:=PLeaf(R[0])^.LText[J]
                           end;
                     // Copy R to left part and cut off right part
                     S:=R; PLeaf(S[0])^.LLen:=I;
                     // Concatenate left and right part into new R
                     MakeConcat(R,Load(S),Load(T),2)
                     end
               else if 0=I then
                  begin
                  // Split at left end keeping track of references
                  S[0]:=Nil; T:=Load(R)
                  end
               else if I=PLeaf(R[0])^.LLen then
                  begin
                  // Split leaf at right end keeping track of references
                  S:=Load(R); T[0]:=Nil
                  end
               else
                  raise ERangeError.Create('Rope index out of bounds');
            $80000000:
               // Rope concatenation
               with PConcat(R[0])^ do
                  // Check weight aginst split point
                  if NWeight<>I then
                     // No exact match, determine left or right branch
                     if I<NWeight then
                        begin
                        // Split left branch
                        Split(Left,I,S,T);
                        // Adjust level
                        L:=Depth(Load(Left));
                        if L>=NLevel then NLevel:=L+1;
                        // Concatenate right branch to result
                        T:=T+Load(Right)
                        end
                     else
                        begin
                        // Split right branch
                        Split(Right,I-NWeight,S,T);
                        // Adjust level
                        L:=Depth(Load(Right));
                        if L>=NLevel then NLevel:=L+1;
                        // Concatenate left branch to result
                        S:=Load(Left)+S
                        end
                  else
                     begin
                     // Exact match, return left and right ropes
                     S:=Load(Left); T:=Load(Right)
                     end
            end
      else
         begin
         // Empty result ropes
         S[0]:=Nil; T[0]:=Nil;
         // Empty rope
         if I<>0 then
            raise ERangeError.Create('Rope index out of bounds')
         end
      end; {Split(var R:Rope; I:Cardinal; var S, T:Rope)}

      // --- Functional core routines

      // Create empty rope variable

      procedure Make(var R:Rope);

      begin {Make(var R:Rope)}
      // Check for rope variable counter overflow
      if RCount=$FFFFFFFF then
         raise ERangeError.Create('Rope variable counter overflow');
      // Increment rope variable count
      RCount:=RCount+1;
      // Create empty rope
      R[0]:=Nil
      end; {Make(var R:Rope)}

      // Drop rope variable

      procedure Drop(var R:Rope);

         var

            L                       : Cardinal;

      begin {Drop(var R:Rope)}
      // Check for rope varible counter zero
      if RCount=$00000000 then
         raise ERangeError.Create('Rope variable counter zero');
      // Decrement rope variable count
      RCount:=RCount-1;
      // Compute rope length
      L:=Length(Load(R));
      // Check for total rope length underflow
      if RLength<L then
         raise ERangeError.Create('Total rope length underflow');
      // Compute new total rope length
      RLength:=RLength-L;
      // Release variable
      Release(R);
      // Leave empty rope
      R[0]:=Nil
      end; {Drop(var R:Rope)}

      // Declare rope value parameter

      procedure Declare(var R:Rope);

         var

            L                       : Cardinal;

      begin {Declare(var R:Rope)}
      // Check for rope variable counter overflow
      if RCount=$FFFFFFFF then
         raise ERangeError.Create('Rope variable counter overflow');
      // Increment rope variable count
      RCount:=RCount+1;
      // Compute rope length
      L:=Length(Load(R));
      // Check for total rope length overflow
      if RLength>$FFFFFFFF-L then
         raise ERangeError.Create('Total rope length overflow');
      // Compute new total rope length
      RLength:=RLength+L
      end; {Declare(var R:Rope)}

      // Pass function result

      procedure Pass(var R:Rope);

         var

            L                       : Cardinal;

      begin {Pass(var R:Rope)}
      // Check for rope variable counter zero
      if RCount=$00000000 then
         raise ERangeError.Create('Rope variable counter zero');
      // Decrement rope variable count
      RCount:=RCount-1;
      // Compute rope length
      L:=Length(Load(R));
      // Check for total rope length underflow
      if RLength<L then
         raise ERangeError.Create('Total rope length underflow');
      // Compute new total rope length
      RLength:=RLength-L
      end; {Pass(var R:Rope)}

      // Allocate rope variable and retrieve value

      function Load(var R:Rope):Rope;

      begin {Load(var R:Rope):Rope}
      // Allocate rope variable and return value
      Load:=Allocate(R)
      end; {Load(var R:Rope):Rope}

      // Create rope value from ANSI string

      function Load(Str:AnsiString):Rope;

         var

            R, S                    : Rope;

            I                       : Cardinal;

      begin {Load(Str:AnsiString):Rope}
      // Check for empty string
      if Length(Str)<>0 then
         begin
         // Clear result
         R[0]:=Nil;
         // Create new leaf
         MakeLeaf(S);
         for I:=1 to Length(Str) do
            begin
            // Copy next character from string
            with PLeaf(S[0])^ do
               begin
               LLen:=LLen+1;
               LText[LLen]:=Str[I];
               end;
            // Check if leaf is full
            if (I mod MaxLLen)=0 then
               begin
               // Concatenate filled leaf to result
               R:=R+S;
               // Continue with next leaf
               if I<>Length(Str) then MakeLeaf(S)
               end
            end;
         // Check for partially filled leaf
         if (I mod MaxLLen)<>0 then R:=R+S;
         // Return result
         Load:=R
         end
      else
         // Empty string
         // Return empty rope
         Load[0]:=Nil
      end; {Load(Str:AnsiString):Rope}

      // Release rope variable and store new value

      procedure Store(var R:Rope; S:Rope);

         var

            L                       : Cardinal;

      begin {Store(var R:Rope; S:Rope)}
      // Compute rope R length
      L:=Length(Load(R));
      // Check for total rope length underflow
      if RLength<L then
         raise ERangeError.Create('Total rope length underflow');
      // Compute new total rope length
      RLength:=RLength-L;
      // Release rope variable
      Release(R);
      // Compute rope S length
      L:=Length(Load(S));
      // Check for total rope length overflow
      if RLength>$FFFFFFFF-L then
         raise ERangeError.Create('Total rope length overflow');
      // Compute new total rope length
      RLength:=RLength+L;
      // Assign new value
      R:=S
      end; {Store(var R:Rope; S:Rope)}

      // Select and return value of R[I]

      function Select(R:Rope; I:Cardinal):Char;

         var

            S                       : Rope;

            Found                   : Boolean;

      begin {Select(R:Rope; I:Cardinal):Char}
      // Initialize
      Found:=False; S:=R;
      // Traverse rope to leaf
      while not Found and (S[0]<>Nil) do
         case S[0]^.TRCount and $80000000 of
            $00000000:
               with PLeaf(S[0])^ do
                  begin
                  // Check index against bounds
                  Found:=(0<I) and (I<=LLen);
                  if Found then
                     // Return item at index
                     Select:=LText[I]
                  else
                     raise ERangeError.Create('Rope index out of bounds')
                  end;
            $80000000:
               with PConcat(S[0])^ do
                  // Select left or right branch
                  if I<=NWeight then
                     // Follow left branch
                     S:=Left
                  else
                     // Adjust index and follow right branch
                     begin
                     I:=I-NWeight; S:=Right
                     end
            end;
      // Check for index out of bounds
      if not Found then
         raise ERangeError.Create('Rope index out of bounds');
      // Release value parameter
      Release(R)
      end; {Select(R:Rope; I:Cardinal):Char}

      // Concatenate ropes S and T

      operator + (S, T:Rope) R:Rope;

         var

            Ds, Dt                  : Level;

            I                       : 1..MaxLLen;

      begin {+ (S, T:Rope) R:Rope}
      // Compute depth of left and right argument
      Ds:=Depth(Load(S)); Dt:=Depth(Load(T));
      if Ds<>Dt then
         if Ds<Dt then
            // Ds < Dt
            case T[0]^.TRCount and $80000000 of
               $00000000:
                  // S is empty rope
                  // T is rope leaf
                  // Return T
                  R:=T;
               $80000000:
                  // T is rope concatenation
                  if Ds<>0 then
                     begin
                     with PConcat(T[0])^ do
                        begin
                        // Create new concatenation node
                        MakeConcat(R,Empty,Load(Right),1);
                        // Recursive descent
                        Release(PConcat(R[0])^.Left);
                        PConcat(R[0])^.Left:=S+Load(Left)
                        end;
                     // Release intermediate value
                     Release(T);
                     with PConcat(R[0])^ do
                        begin
                        // Adjust weight in left branch
                        NWeight:=Length(Load(Left));
                        // Compute left and right levels
                        Ds:=Depth(Load(Left)); Dt:=Depth(Load(Right));
                        // Adjust level
                        if Ds<Dt then NLevel:=Dt+1 else NLevel:=Ds+1
                        end;
                     // Rebalance result
                     Rebalance(R)
                     end
                  else
                     // S is empty rope
                     // Return T
                     R:=T
               end
         else
            // Ds > Dt
            case S[0]^.TRCount and $80000000 of
               $00000000:
                  // S is rope leaf
                  // T is empty rope
                  // Return S
                  R:=S;
               $80000000:
                  // S is rope concatenation
                  if Dt<>0 then
                     begin
                     with PConcat(S[0])^ do
                        begin
                        // Create new concatenation node
                        MakeConcat(R,Load(Left),Empty,1);
                        // Recursive descent
                        Release(PConcat(R[0])^.Right);
                        PConcat(R[0])^.Right:=Load(Right)+T
                        end;
                     // Release intermediate value
                     Release(S);
                     with PConcat(R[0])^ do
                        begin
                        // No weight adjustment in right branch
                        // Compute left and right levels
                        Ds:=Depth(Load(Left)); Dt:=Depth(Load(Right));
                        // Adjust level
                        if Ds<Dt then NLevel:=Dt+1 else NLevel:=Ds+1
                        end;
                     // Rebalance result
                     Rebalance(R)
                     end
                  else
                     // T is empty rope
                     // Return S
                     R:=S
               end
      else
         // Ds = Dt
         if S[0]<>Nil { implies T[0]<>Nil } then
            case S[0]^.TRCount and $80000000 of
               $00000000:
                  // Check if leaves fit in a single leaf
                  if Length(Load(S))+Length(Load(T))<=MaxLLen then
                     begin
                     // Create new mutable leaf
                     MakeLeaf(R);
                     with PLeaf(R[0])^ do
                        begin
                        // Copy text S
                        for I:=1 to PLeaf(S[0])^.LLen do
                           begin
                           LLen:=LLen+1;
                           LText[LLen]:=PLeaf(S[0])^.LText[I]
                           end;
                        // Copy text T
                        for I:=1 to PLeaf(T[0])^.LLen do
                           begin
                           LLen:=LLen+1;
                           LText[LLen]:=PLeaf(T[0])^.LText[I]
                           end
                        end;
                     // Release intermediate ropes
                     Release(S); Release(T)
                     end
                  else
                     // Leaves do not fit within a single leaf
                     // Return concatenation of leaf nodes
                     MakeConcat(R,S,T,Ds+1 { = Dt+1 } );
               $80000000:
                  // Return concatenation of concatenation nodes
                  MakeConcat(R,S,T,Ds+1 { = Dt+1 } )
               end
         else
            // S and T both empty
            R[0]:=Nil
      end; {+ (S, T:Rope) R:Rope}

      // Return true when rope S equal T

      operator = (S, T:Rope) R:Boolean;

         var

            Ls, Lt, I               : Cardinal;

      begin {= (S, T:Rope) R:Boolean}
      // Check for root node equality
      if S[0]<>T[0] then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         if Ls=Lt then
            begin
            // S and T have equal length
            I:=1;
            // Compare characters
            while (I<=Ls) { implies (I<Lt) } and
               (Select(Load(S),I)=Select(Load(T),I)) do I:=I+1;
            // Set result
            R:=I>Ls
            end
         else
            // S and T have different length
            R:=False
         end
      else
         // S and T refer to the same rope
         R:=True;
      // Release value parameters
      Release(S); Release(T)
      end; {= (S, T:Rope) R:Boolean}

      // Return true when rope S unequal T

      operator <> (S, T:Rope) R:Boolean;

         var

            Ls, Lt, I               : Cardinal;

      begin {<> (S, T:Rope) R:Boolean}
      // Check for root node equality
      if S[0]<>T[0] then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         if Ls=Lt then
            begin
            // S and T have equal length
            I:=1;
            // Compare characters
            while (I<=Ls) { implies (I<Lt) } and
               (Select(Load(S),I)=Select(Load(T),I)) do I:=I+1;
            // Set result
            R:=I<=Ls
            end
         else
            // S and T have different length
            R:=True
         end
      else
         // S and T refer to the same rope
         R:=False;
      // Release value parameters
      Release(S); Release(T)
      end; {<> (S, T:Rope) R:Boolean}

      // Return true when rope S strictly less than T

      operator < (S, T:Rope) R:Boolean;

         var

            Ls, Lt, I               : Cardinal;

      begin {< (S, T:Rope) R:Boolean}
      // Check for root node equality
      if S[0]<>T[0] then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         // Skip characters that are equal
         I:=1;
         while (I<=Ls) and (I<=Lt) and
            (Select(Load(S),I)=Select(Load(T),I)) do I:=I+1;
         // Check if unequal item was found
         if (I<=Ls) and (I<=Lt) then
            // Compare characters if so
            R:=Select(Load(S),I)<Select(Load(T),I)
         else
            // Set result when Ls<Lt
            R:=Ls<Lt
         end
      else
         // S and T refer to the same rope
         R:=False;
      // Release value parameters
      Release(S); Release(T)
      end; {< (S, T:Rope) R:Boolean}

      // Return true when rope S strictly greater than T

      operator > (S, T:Rope) R:Boolean;

         var

            Ls, Lt, I               : Cardinal;

      begin {> (S, T:Rope) R:Boolean}
      // Check for root node equality
      if S[0]<>T[0] then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         // Skip characters that are equal
         I:=1;
         while (I<=Ls) and (I<=Lt) and
            (Select(Load(S),I)=Select(Load(T),I)) do I:=I+1;
         // Check if unequal item was found
         if (I<=Ls) and (I<=Lt) then
            // Compare characters if so
            R:=Select(Load(S),I)>Select(Load(T),I)
         else
            // Set result when Ls<Lt
            R:=Ls>Lt
         end
      else
         // S and T refer to the same rope
         R:=False;
      // Release value parameters
      Release(S); Release(T)
      end; {> (S, T:Rope) R:Boolean}

      // Return true when rope S less than or equal T

      operator <= (S, T:Rope) R:Boolean;

         var

            Ls, Lt, I               : Cardinal;

      begin {<= (S, T:Rope) R:Boolean}
      // Check for root node equality
      if S[0]<>T[0] then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         // Skip characters that are equal
         I:=1;
         while (I<=Ls) and (I<=Lt) and
            (Select(Load(S),I)=Select(Load(T),I)) do I:=I+1;
         // Check if unequal item was found
         if (I<=Ls) and (I<=Lt) then
            // Compare characters if so
            R:=Select(Load(S),I)<=Select(Load(T),I)
         else
            // Set result when Ls<Lt
            R:=Ls<=Lt
         end
      else
         // S and T refer to the same rope
         R:=True;
      // Release value parameters
      Release(S); Release(T)
      end; {<= (S, T:Rope) R:Boolean}

      // Return true when rope S greater than or equal T

      operator >= (S, T:Rope) R:Boolean;

         var

            Ls, Lt, I               : Cardinal;

      begin {>= (S, T:Rope) R:Boolean}
      // Check for root node equality
      if S[0]<>T[0] then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         // Skip characters that are equal
         I:=1;
         while (I<=Ls) and (I<=Lt) and
            (Select(Load(S),I)=Select(Load(T),I)) do I:=I+1;
         // Check if unequal item was found
         if (I<=Ls) and (I<=Lt) then
            // Compare characters if so
            R:=Select(Load(S),I)>=Select(Load(T),I)
         else
            // Set result when Ls<Lt
            R:=Ls>=Lt
         end
      else
         // S and T refer to the same rope
         R:=True;
      // Release value parameters
      Release(S); Release(T)
      end; {>= (S, T:Rope) R:Boolean}

      // Return length of rope R

      function Length(R:Rope):Cardinal;

      begin {Length(R:Rope):Cardinal}
      // Check for empty rope
      if R[0]<>Nil then
         case R[0]^.TRCount and $80000000 of
            $00000000:
               // Rope leaf
               Length:=PLeaf(R[0])^.LLen;
            $80000000:
               // Rope concatenation
               with PConcat(R[0])^ do
                  Length:=NWeight+Length(Load(Right))
            end
      else
         // Empty rope
         Length:=0;
      // Release value parameter
      Release(R)
      end; {Length(R:Rope):Cardinal}

      // Return length of string S

      function Length(S:AnsiString):Cardinal;

      begin {Length(S:AnsiString):Cardinal}
      // Function is needed to allow normal use of the overloaded function
      // Just call the system function and return result
      Length:=System.Length(S)
      end; {Length(S:AnsiString):Cardinal}

      // Return position of first occurrence of rope S in rope T

      function Pos(S, T:Rope):Cardinal;

         var

            Ls, Lt, I, J            : Cardinal;

      begin {Pos(S, T:Rope):Cardinal}
      // Check for empty rope
      if S[0]<>Nil then
         begin
         // Compare length of S and T
         Ls:=Length(Load(S)); Lt:=Length(Load(T));
         if Ls<=Lt then
            begin
            // Initialize result
            Pos:=0;
            // Search S within T
            I:=1; J:=1;
            while (I+Ls-1<=Lt) and (J<=Ls) do
               begin
               // Compare S[J] with T[I+J-1]
               while (J<=Ls) and
                  (Select(Load(S),J)=Select(Load(T),I+J-1)) do J:=J+1;
               // Check if found
               if J>Ls then
                  // S found at position T[I]
                  Pos:=I
               else
                  begin
                  // Continue search
                  I:=I+1; J:=1
                  end
               end
            end
         else
            // Length S is greater than length T
            Pos:=0
         end
      else
         // S is empty rope
         Pos:=0;
      // Release value parameters
      Release(S); Release(T)
      end; {Pos(S, T:Rope):Cardinal}

      // Concatenate ropes R1 and R2

      function Concat(R1, R2:Rope):Rope;

      begin {Concat(R1, R2:Rope):Rope}
      // Concatenate ropes
      Concat:=R1+R2
      end; {Concat(R1, R2:Rope):Rope}

      // Concatenate ropes R1, R2 and R3

      function Concat(R1, R2, R3:Rope):Rope;

      begin {Concat(R1, R2, R3:Rope):Rope}
      // Concatenate ropes
      Concat:=R1+R2+R3
      end; {Concat(R1, R2, R3:Rope):Rope}

      // Concatenate ropes R1, R2, R3 and R4

      function Concat(R1, R2, R3, R4:Rope):Rope;

      begin {Concat(R1, R2, R3, R4:Rope):Rope}
      // Concatenate ropes
      Concat:=R1+R2+R3+R4
      end; {Concat(R1, R2, R3, R4:Rope):Rope}

      // Return rope of L characters from R starting at R[I]

      function Copy(R:Rope; I, L:Cardinal):Rope;

         var

            S, T, U                 : Rope;

      begin {Copy(R:Rope; I, L:Cardinal):Rope}
      // Check index
      if I<=Length(Load(R)) then
         begin
         // Split R at index
         Split(R,Cardinal(I-1),S,T);
         // Release left part
         Release(S);
         // Check length
         if L<=Length(Load(T)) then
            begin
            // Split right part at desired length
            Split(T,L,U,S);
            // Release right part
            Release(S);
            // Return result
            Release(T); Copy:=U
            end
         else
            // Selection past end
            // Return right part
            Copy:=T
         end
      else
         // Index past end
         // Return empty rope
         Copy[0]:=Nil;
      // Release value parameters
      Release(R)
      end; {Copy(R:Rope; I, L:Cardinal):Rope}

      // Delete L characters from R starting at R[I]

      procedure Delete(var R:Rope; I, L:Cardinal);

         var

            S, T, U, V              : Rope;

            Lu                      : Cardinal;

      begin {Delete(var R:Rope; I, L:Cardinal)}
      // Check index
      if I<=Length(Load(R)) then
         begin
         // Split R at index
         Split(R,Cardinal(I-1),S,U);
         // Release split rope
         Release(R);
         // Compute length right part
         Lu:=Length(Load(U));
         // Check length
         if L<=Lu then
            begin
            // Split right part at desired length
            Split(U,L,V,T);
            // Release intermediate values
            Release(U); Release(V);
            // Check for total rope length underflow
            if RLength<L then
               raise ERangeError.Create('Total rope length underflow');
            // Compute new total rope length
            RLength:=RLength-L;
            // Keep left and right part
            R:=S+T
            end
         else
            // Selection past end
            begin
            // Check for total rope length underflow
            if RLength<Lu then
               raise ERangeError.Create('Total rope length underflow');
            // Compute new total rope length
            RLength:=RLength-Lu;
            // Release intermediate value
            Release(U);
            // Keep left part
            R:=S
            end
         end
      end; {Delete(var R:Rope; I, L:Cardinal)}

      // Insert rope S in rope R after R[I]

      procedure Insert(S:Rope; var R:Rope; I:Cardinal);

         var

            T, U, V                 : Rope;

            L                       : Cardinal;

      begin {Insert(S:Rope; var R:Rope; I:Cardinal)}
      // Compute rope S length
      L:=Length(Load(S));
      // Check for total rope length overflow
      if RLength>$FFFFFFFF-L then
         raise ERangeError.Create('Total rope length overflow');
      // Compute new total rope length
      RLength:=RLength+L;
      // Check index
      if I<=Length(Load(R)) then
         begin
         // Split R after index
         Split(R,I,T,U);
         // Concatenate and release parts
         Release(R); R:=T+S+U
         end
      else
         // Selection past end
         // Append and release S
         R:=R+S
      end; {Insert(S:Rope; var R:Rope; I:Cardinal)}

      // Return conversion of R to upper case

      function Upcase(R:Rope):Rope;

         var

            S                       : Rope;

            Ch                      : Char;

      begin {Upcase(R:Rope):Rope}
      // Clear result
      S[0]:=Nil;
      // Loop through R appending to S
      for Ch in Load(R) do
         // Check for lower case characters
         if Ch in ['a'..'z'] then
            // Convert to upper case and append to S
            S:=S+Load(Chr(Ord(Ch)-Ord('a')+Ord('A')))
         else
            // Append to S
            S:=S+Load(Ch);
      // Return S
      Upcase:=S;
      // Release value parameter
      Release(R)
      end; {Upcase(R:Rope):Rope}

      // Return conversion of R to lower case

      function Lowercase(R:Rope):Rope;

         var

            S                       : Rope;

            Ch                      : Char;

      begin {Lowercase(R:Rope):Rope}
      // Clear result
      S[0]:=Nil;
      // Loop through R appending to S
      for Ch in Load(R) do
         // Check for upper case characters
         if Ch in ['A'..'Z'] then
            // Convert to lower case and append to S
            S:=S+Load(Chr(Ord(Ch)-Ord('A')+Ord('a')))
         else
            // Append to S
            S:=S+Load(Ch);
      // Return S
      Lowercase:=S;
      // Release value parameter
      Release(R)
      end; {Lowercase(R:Rope):Rope}

      // Read rope text from standard input

      procedure ReadRope(var R:Rope);

         var

            Ch                      : Char;

      begin {ReadRope(var R:Rope)}
      // Clear result
      Store(R,Empty);
      // Read characters up to end of line
      while not EoLn do
         begin
         // Read character from standard input
         Read(Ch);
         // Append to rope
         Store(R,Load(R)+Load(Ch))
         end
      end; {ReadRope(var R:Rope)}

      // Read rope text and line terminator from standard input

      procedure ReadLnRope(var R:Rope);

         var

            Ch                      : Char;

      begin {ReadLnRope(var R:Rope)}
      // Clear result
      Store(R,Empty);
      // Read characters up to end of line
      while not EoLn do
         begin
         // Read character from standard input
         Read(Ch);
         // Append to rope
         Store(R,Load(R)+Load(Ch))
         end;
      // Read past end of line
      ReadLn
      end; {ReadLnRope(var R:Rope)}

      // Read rope text from text file

      procedure ReadRope(var T:Text; var R:Rope);

         var

            Ch                      : Char;

      begin {ReadRope(var T:Text; var R:Rope)}
      // Clear result
      Store(R,Empty);
      // Read characters up to end of line
      while not EoLn(T) do
         begin
         // Read character from text file
         Read(T,Ch);
         // Append to rope
         Store(R,Load(R)+Load(Ch))
         end
      end; {ReadRope(var T:Text; var R:Rope)}

      // Read rope text and line terminator from text file

      procedure ReadLnRope(var T:Text; var R:Rope);

         var

            Ch                      : Char;

      begin {ReadLnRope(var T:Text; var R:Rope)}
      // Clear result
      Store(R,Empty);
      // Read characters up to end of line
      while not EoLn(T) do
         begin
         // Read character from text file
         Read(T,Ch);
         // Append to rope
         Store(R,Load(R)+Load(Ch))
         end;
      // Read past end of line
      ReadLn(T)
      end; {ReadLnRope(var T:Text; var R:Rope)}

      // Write rope text to standard output

      procedure WriteRope(R:Rope);

         var

            Ch                      : Char;

      begin {WriteRope(R:Rope)}
      // Write rope to standard output
      for Ch in Load(R) do Write(Ch);
      // Release value parameter
      Release(R)
      end; {WriteRope(R:Rope)}

      // Write rope text and line terminator to standard output

      procedure WriteLnRope(R:Rope);

         var

            Ch                      : Char;

      begin {WriteLnRope(R:Rope)}
      // Write rope to standard output
      for Ch in Load(R) do Write(Ch);
      // Write line terminator
      WriteLn;
      // Release value parameter
      Release(R)
      end; {WriteLnRope(R:Rope)}

      // Write rope text to text file

      procedure WriteRope(var T:Text; R:Rope);

         var

            Ch                      : Char;

      begin {WriteRope(var T:Text; R:Rope)}
      // Write rope to text file
      for Ch in Load(R) do Write(T,Ch);
      // Release value parameter
      Release(R)
      end; {WriteRope(var T:Text; R:Rope)}

      // Write rope text and line terminator to text file

      procedure WriteLnRope(var T:Text; R:Rope);

         var

            Ch                      : Char;

      begin {WriteLnRope(var T:Text; R:Rope)}
      // Write rope to text file
      for Ch in Load(R) do Write(T,Ch);
      // Write line terminator
      WriteLn(T);
      // Release value parameter
      Release(R)
      end; {WriteLnRope(var T:Text; R:Rope)}

      // --- Support and diagnostic routines ---

      // Enumerator operator to allow for .. in loops

      operator enumerator (R:Rope):REnumerator;

      begin {enumerator (R:Rope):REnumerator}
      // Create and return enumerator instance
      Result:=REnumerator.Create(R)
      end; {enumerator (R:Rope):REnumerator}

      // Return rope leaf node count for rope R

      function LeafNodes(R:Rope):Cardinal;

      begin {LeafNodes(R:Rope):Cardinal}
      // Check for empty rope
      if R[0]<>Nil then
         case R[0]^.TRCount and $80000000 of
            $00000000:
               // Leaf node
               LeafNodes:=1;
            $80000000:
               // Concatenation node
               with PConcat(R[0])^ do
                  LeafNodes:=LeafNodes(Load(Left))+LeafNodes(Load(Right))
            end
      else
         // Empty rope
         LeafNodes:=0;
      // Release value parameter
      Release(R)
      end; {LeafNodes(R:Rope):Cardinal}

      // Return rope concatenation node count for rope R

      function ConcatNodes(R:Rope):Cardinal;

      begin {ConcatNodes(R:Rope):Cardinal}
      // Check for empty rope
      if R[0]<>Nil then
         case R[0]^.TRCount and $80000000 of
            $00000000:
               // Leaf nodes
               ConcatNodes:=0;
            $80000000:
               // Concatenation node
               with PConcat(R[0])^ do
                  ConcatNodes:=1+ConcatNodes(Load(Left))
                     +ConcatNodes(Load(Right))
            end
      else
         // Empty rope
         ConcatNodes:=0;
      // Release value parameter
      Release(R)
      end; {ConcatNodes(R:Rope):Cardinal}

      // Return rope tree depth for rope R

      function Depth(R:Rope):Level;

      begin {Depth(R:Rope):Level}
      // Check for empty rope
      if R[0]<>Nil then
         case R[0]^.TRCount and $80000000 of
            $00000000:
               // Rope leaf
               Depth:=1;
            $80000000:
               // Rope concatenation
               Depth:=PConcat(R[0])^.NLevel
            end
      else
         // Empty rope
         Depth:=0;
      // Release value parameter
      Release(R)
      end; {Depth(R:Rope):Level}

      // Return current rope variable count

      function RopeVariables:Cardinal;

      begin {RopeVariables:Cardinal}
      // Return rope variable count
      RopeVariables:=RCount
      end; {RopeVariables:Cardinal}

      // Return total leaf node count for current rope variables

      function LeafNodes:Cardinal;

      begin {LeafNodes:Cardinal}
      // Return leaf code count
      LeafNodes:=LCount
      end; {LeafNodes:Cardinal}

      // Return concatenation node count for current rope variables

      function ConcatNodes:Cardinal;

      begin {ConcatNodes:Cardinal}
      // Return concatenation node count
      ConcatNodes:=CCount
      end; {ConcatNodes:Cardinal}

      // Return total length of current rope variables

      function Length:Cardinal;

      begin {Length:Cardinal}
      // Return total length
      Length:=RLength
      end; {Length:Cardinal}

      // --- Debug routines ---

      {$IF DEFINED(DEBUG)}

      // Test split rope R at index I in S and T

      procedure TestSplit(var R:Rope; I:Cardinal; var S, T:Rope);

         var

            L                       : Cardinal;

      begin {TestSplit(var R:Rope; I:Cardinal; var S, T:Rope)}
      // Compute rope R length
      L:=Length(Load(R));
      // Check for total rope length overflow
      if RLength>$FFFFFFFF-L then
         raise ERangeError.Create('Total rope length overflow');
      // Compute new total rope length
      RLength:=RLength+L;
      // Perform rope split
      Split(R,I,S,T)
      end; {TestSplit(var R:Rope; I:Cardinal; var S, T:Rope)}

      // List structure of rope

      procedure List(R:Rope);

         var

            I                       : 0..MaxLLen;

      begin {List(R:Rope)}
      // Check for empty rope
      if R[0]<>Nil then
         case R[0]^.TRCount and $80000000 of
            $00000000:
               // Rope leaf
               with PLeaf(R[0])^ do
                  begin
                  // List leaf node contents
                  Write((TRCount and $7FFFFFFF)-1,',',LLen,':''');
                  for I:=1 to LLen do Write(LText[I]);
                  Write('''')
                  end;
            $80000000:
               // Rope concatenation
               with PConcat(R[0])^ do
                  begin
                  // List concatenation node and left and right branches
                  Write((TRCount and $7FFFFFFF)-1,',',NLevel,',',NWeight);
                  Write(':(');
                  List(Load(Left)); Write(','); List(Load(Right));
                  Write(')')
                  end
               end
      else
         // Empty rope
         Write('.');
      // Release value parameter
      Release(R)
      end; {List(R:Rope)}

      {$ENDIF}

   end. {Ropes}
